# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""User profiles module for Invenio."""

from flask_menu import current_menu
from invenio_i18n import LazyString
from invenio_i18n import lazy_gettext as _
from invenio_theme.proxies import current_theme_icons

from . import config
from .api import current_userprofile
from .forms import confirm_register_form_factory, register_form_factory


class InvenioUserProfiles(object):
    """Invenio-UserProfiles extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)

        # Register current_profile
        app.context_processor(lambda: dict(current_userprofile=current_userprofile))

        app.extensions["invenio-userprofiles"] = self

    def init_config(self, app):
        """Initialize configuration."""
        excludes = [
            "USERPROFILES_BASE_TEMPLATE",
            "USERPROFILES_SETTINGS_TEMPLATE",
        ]
        for k in dir(config):
            if k.startswith("USERPROFILES_") and k not in excludes:
                app.config.setdefault(k, getattr(config, k))

        app.config.setdefault("USERPROFILES", True)

        app.config.setdefault(
            "USERPROFILES_BASE_TEMPLATE",
            app.config.get("BASE_TEMPLATE", "invenio_userprofiles/base.html"),
        )

        app.config.setdefault(
            "USERPROFILES_SETTINGS_TEMPLATE",
            app.config.get(
                "SETTINGS_TEMPLATE", "invenio_userprofiles/settings/base.html"
            ),
        )

        if app.config["USERPROFILES_EXTEND_SECURITY_FORMS"]:
            app.config.setdefault(
                "USERPROFILES_REGISTER_USER_BASE_TEMPLATE",
                app.config.get(
                    "SECURITY_REGISTER_USER_TEMPLATE",
                    "invenio_accounts/register_user.html",
                ),
            )

            app.config["SECURITY_REGISTER_USER_TEMPLATE"] = (
                "invenio_userprofiles/register_user.html"
            )


def finalize_app(app):
    """Finalize app.

    NOTE: replace former @record_once decorator
    """
    init_common(app)
    init_menu(app)


def api_finalize_app(app):
    """Finalize app for api.

    NOTE: replace former @record_once decorator
    """
    init_common(app)


def init_common(app):
    """Post initialization."""
    if app.config["USERPROFILES_EXTEND_SECURITY_FORMS"]:
        security_ext = app.extensions["security"]
        security_ext.confirm_register_form = confirm_register_form_factory(
            security_ext.confirm_register_form
        )
        security_ext.register_form = register_form_factory(security_ext.register_form)


def init_menu(app):
    """Init menu."""
    current_menu.submenu("settings.profile").register(
        endpoint="invenio_userprofiles.profile",
        text=_(
            "%(icon)s Profile",
            icon=LazyString(lambda: f'<i class="{current_theme_icons.user}"></i>'),
        ),
        order=0,
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2022 CERN.
# Copyright (C) 2021      TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""API for user profiles."""

from warnings import warn

from flask_security import current_user
from werkzeug.local import LocalProxy

from .models import UserProfileProxy


def _get_current_userprofile():
    """Get current user profile.

    .. note:: If the user is anonymous, then a
        :class:`invenio_userprofiles.models.AnonymousUserProfile` instance is
        returned.

    :returns: The :class:`invenio_userprofiles.models.UserProfile` instance.
    """
    warn(
        "current_userprofile is deprecated, use current_user instead",
        DeprecationWarning,
    )
    return UserProfileProxy(current_user)


current_userprofile = LocalProxy(lambda: _get_current_userprofile())
"""Proxy to the user profile of the currently logged in user."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Forms for user profiles."""

import pytz
from flask import current_app
from flask_login import current_user
from flask_security.forms import email_required, email_validator, unique_user_email
from flask_wtf import FlaskForm
from invenio_i18n import lazy_gettext as _
from invenio_i18n.ext import InvenioI18N
from werkzeug.local import LocalProxy
from wtforms import (
    FormField,
    RadioField,
    SelectField,
    StringField,
    SubmitField,
    validators,
)
from wtforms.validators import (
    DataRequired,
    EqualTo,
    Length,
    StopValidation,
    ValidationError,
)

from .models import UserProfileProxy
from .validators import USERNAME_RULES, validate_username


def strip_filter(text):
    """Filter for trimming whitespace.

    :param text: The text to strip.
    :returns: The stripped text.
    """
    return text.strip() if text else text


def current_user_email(form, field):
    """Field validator to stop validation if email wasn't changed."""
    if current_user.email.lower() == field.data.lower():
        raise StopValidation()


class ProfileForm(FlaskForm):
    """Form for editing user profile."""

    profile_proxy_cls = UserProfileProxy

    username = StringField(
        # NOTE: Form field label
        _("Username"),
        # NOTE: Form field help text
        description=_("Required. %(username_rules)s", username_rules=USERNAME_RULES),
        validators=[Length(max=50), DataRequired(message=_("Username not provided."))],
        filters=[strip_filter],
    )

    full_name = StringField(
        # NOTE: Form label
        _("Full name"),
        validators=[Length(max=255)],
        filters=[strip_filter],
    )

    affiliations = StringField(
        # NOTE: Form label
        _("Affiliations"),
        validators=[Length(max=255)],
        filters=[strip_filter],
    )

    def validate_username(self, field):
        """Wrap username validator for WTForms."""
        try:
            validate_username(field.data)
        except ValueError as e:
            raise ValidationError(e)

        # Check if username is already taken.
        datastore = current_app.extensions["security"].datastore
        user = datastore.find_user(username=field.data)
        if user is None:
            return

        # NOTE: Form validation error.
        msg = _("Username is not available.")

        if current_user.is_anonymous:
            # We are handling a new sign up (i.e. anonymous user) AND a
            # the username already exists. Fail.
            raise ValidationError(msg)
        else:
            # We are handling a user editing their profile AND a
            # the username already exists.
            is_same_user = current_user.get_id() == str(user.id)
            if not is_same_user:
                # Username already taken by another user.
                raise ValidationError(msg)

    def process(self, formdata=None, obj=None, data=None, extra_filters=None, **kwargs):
        """Build a proxy around the object."""
        if obj is not None:
            obj = self.profile_proxy_cls(obj)
        super().process(
            formdata=formdata, obj=obj, data=data, extra_filters=extra_filters, **kwargs
        )

    def populate_obj(self, user):
        """Populates the obj."""
        user = self.profile_proxy_cls(user)
        super().populate_obj(user)


class EmailProfileForm(ProfileForm):
    """Form to allow editing of email address."""

    email = StringField(
        # NOTE: Form field label
        _("Email address"),
        filters=[
            lambda x: x.lower() if x is not None else x,
        ],
        validators=[
            email_required,
            current_user_email,
            email_validator,
            unique_user_email,
        ],
    )

    email_repeat = StringField(
        # NOTE: Form field label
        _("Re-enter email address"),
        # NOTE: Form field help text
        description=_("Please re-enter your email address."),
        filters=[
            lambda x: x.lower() if x else x,
        ],
        validators=[
            email_required,
            # NOTE: Form validation error.
            EqualTo("email", message=_("Email addresses do not match.")),
        ],
    )


class VerificationForm(FlaskForm):
    """Form to render a button to request email confirmation."""

    # NOTE: Form button label
    send_verification_email = SubmitField(_("Resend verification email"))


def register_form_factory(Form):
    """Factory for creating an extended user registration form."""

    class CsrfDisabledProfileForm(ProfileForm):
        """Subclass of ProfileForm to disable CSRF token in the inner form.

        This class will always be a inner form field of the parent class
        `Form`. The parent will add/remove the CSRF token in the form.
        """

        def __init__(self, *args, **kwargs):
            """Initialize the object by hardcoding CSRF token to false."""
            kwargs = _update_with_csrf_disabled(kwargs)
            super(CsrfDisabledProfileForm, self).__init__(*args, **kwargs)

    class RegisterForm(Form):
        """RegisterForm extended with UserProfile details."""

        # cannot be called `user_profile`, to avoid naming collision with the
        # hybrid property in the model
        profile = FormField(CsrfDisabledProfileForm, separator=".")

        def to_dict(self):
            profile_data = self.profile.data
            data = super().to_dict()
            data["username"] = profile_data.pop("username")
            data["user_profile"] = profile_data
            return data

    return RegisterForm


class PreferencesForm(FlaskForm):
    """Form for editing user profile."""

    profile_proxy_cls = UserProfileProxy

    visibility = RadioField(
        _("Profile visibility"),
        choices=[
            ("public", _("Public")),
            ("restricted", _("Hidden")),
        ],
        description=_(
            "Public profiles can be found by other users via searches on "
            "username, full name and affiliation. Hidden profiles cannot be"
            " found by other users."
        ),
    )

    email_visibility = RadioField(
        _("Email visibility"),
        choices=[
            ("public", _("Public")),
            ("restricted", _("Hidden")),
        ],
        description=_(
            "Public email visibility enables your profile to be found by "
            "your email address."
        ),
    )

    locale = LocalProxy(
        lambda: SelectField(
            _("Preferences locale"),
            choices=set(current_app.extensions["invenio-i18n"].get_locales()),
        ),
    )

    # timezone = SelectField(
    #    _("Preferences timezone"),
    #    choices=pytz.all_timezones,
    #    validators=[validators.InputRequired()],
    # )

    def process(self, formdata=None, obj=None, data=None, extra_filters=None, **kwargs):
        """Build a proxy around the object."""
        if obj is not None:
            obj = self.profile_proxy_cls(obj)
        super().process(
            formdata=formdata, obj=obj, data=data, extra_filters=extra_filters, **kwargs
        )

    def populate_obj(self, user):
        """Populates the obj."""
        user = self.profile_proxy_cls(user)
        super().populate_obj(user)


def confirm_register_form_factory(Form):
    """Factory for creating a confirm register form with UserProfile fields."""

    class CsrfDisabledProfileForm(ProfileForm):
        """Subclass of ProfileForm to disable CSRF token in the inner form.

        This class will always be an inner form field of the parent class
        `Form`. The parent will add/remove the CSRF token in the form.
        """

        def __init__(self, *args, **kwargs):
            """Initialize the object by hardcoding CSRF token to false."""
            kwargs = _update_with_csrf_disabled(kwargs)
            super().__init__(*args, **kwargs)

    class ConfirmRegisterForm(Form):
        """RegisterForm extended with UserProfile details."""

        # cannot be called `user_profile`, to avoid naming collision with the
        # hybrid property in the model
        profile = FormField(CsrfDisabledProfileForm, separator=".")

        def to_dict(self):
            profile_data = self.profile.data
            data = super().to_dict()
            data["username"] = profile_data.pop("username")
            data["user_profile"] = profile_data
            return data

    return ConfirmRegisterForm


def confirm_register_form_preferences_factory(Form):
    """Factory for creating a confirm register form with UserProfile and Preferences."""

    class CsrfDisabledPreferencesForm(PreferencesForm):
        """Subclass of PreferencesForm to disable CSRF token in the inner form."""

        def __init__(self, *args, **kwargs):
            """Initialize the object by hardcoding CSRF token to false."""
            kwargs = _update_with_csrf_disabled(kwargs)
            super().__init__(*args, **kwargs)

    class ConfirmRegisterForm(Form):
        """RegisterForm extended with Preferences details."""

        # cannot be called `preferences`, to avoid naming collision with the
        # hybrid property in the model
        prefs = FormField(CsrfDisabledPreferencesForm, separator=".")

        def to_dict(self):
            preferences_data = self.prefs.data
            data = super().to_dict()
            data["preferences"] = preferences_data
            return data

    return ConfirmRegisterForm


def _update_with_csrf_disabled(d=None):
    """Update the input dict with CSRF disabled."""
    if d is None:
        d = {}
    d.setdefault("meta", {})
    d["meta"].update({"csrf": False})
    return d



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Database models for user profiles."""

from flask import current_app


class AnonymousUserProfile:
    """Anonymous user profile."""

    @property
    def is_anonymous(self):
        """Return whether this UserProfile is anonymous."""
        return True


class UserProfileProxy:
    """Proxy for a user that allows mapping the form to the user object."""

    _profile_attrs = ["full_name", "affiliations"]
    _preferences_attrs = ["email_visibility", "visibility", "locale", "timezone"]
    _read_only_attrs = ["email_repeat"]
    _aliases = {"email_repeat": "email", "user_id": "id"}

    def __init__(self, user):
        """."""
        super().__setattr__("_user", user)

    def __getattr__(self, attr):
        """."""
        if attr in self._profile_attrs:
            return self._user.user_profile.get(attr, None)
        elif attr in self._preferences_attrs:
            return self._user.preferences.get(attr, None)
        else:
            attr = self._aliases.get(attr, attr)
            return getattr(self._user, attr)

    def __setattr__(self, attr, value):
        """."""
        if attr == "email":
            if (
                current_app.config["USERPROFILES_EMAIL_ENABLED"]
                and self._user.email != value
            ):
                self._user.email = value
                self._user.confirmed_at = None
        elif attr in self._profile_attrs:
            self._user.user_profile = {**self._user.user_profile, attr: value}
        elif attr in self._preferences_attrs:
            self._user.preferences = {**self._user.preferences, attr: value}
        elif attr not in self._read_only_attrs:
            setattr(self._user, attr, value)

    def __hasattr__(self, attr):
        """."""
        if attr in self._profile_attrs:
            return attr in self._user.user_profile
        elif attr in self._preferences_attrs:
            return attr in self._user.preferences[attr]
        else:
            attr = self._aliases.get(attr, attr)
            hasattr(self._user, attr)

    @classmethod
    def get_by_username(cls, username):
        """Get profile by username.

        :param username: A username to query for (case insensitive).
        """
        # Kept for backward compatibility
        user = current_app.extensions["security"].datastore.find_user(
            _username=username.lower()
        )
        return cls(user) if user else None

    @classmethod
    def get_by_userid(cls, user_id):
        """Get profile by user identifier.

        :param user_id: Identifier of a :class:`~invenio_accounts.models.User`.
        :returns: A :class:`~invenio_userprofiles.models.UserProfile` instance
            or ``None``.
        """
        # Kept for backward compatibility
        user = current_app.extensions["security"].datastore.find_user(id=user_id)
        return cls(user) if user else None


# Backward compatibility
UserProfile = UserProfileProxy



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2024-2025 Graz University of Technology.
# Copyright (C) 2025 KTH Royal Institute of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""User profiles module for Invenio."""

# Monkey patch Werkzeug 2.1
# Flask-Login uses the safe_str_cmp method which has been removed in Werkzeug
# 2.1. Flask-Login v0.6.0 (yet to be released at the time of writing) fixes the
# issue. Once we depend on Flask-Login v0.6.0 as the minimal version in
# Flask-Security-Invenio/Invenio-Accounts we can remove this patch again.
try:
    # Werkzeug <2.1
    from werkzeug import security

    security.safe_str_cmp
except AttributeError:
    # Werkzeug >=2.1
    import hmac

    from werkzeug import security

    security.safe_str_cmp = hmac.compare_digest

from .api import current_userprofile
from .ext import InvenioUserProfiles
from .models import UserProfile, UserProfileProxy

__version__ = "4.1.1"

__all__ = (
    "__version__",
    "current_userprofile",
    "InvenioUserProfiles",
    "UserProfile",
    "UserProfileProxy",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2023-2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that adds userprofiles to the platform."""

from warnings import warn

from flask import (
    Blueprint,
    current_app,
    flash,
    redirect,
    render_template,
    request,
    url_for,
)
from flask_login import current_user, login_required
from flask_security.confirmable import send_confirmation_instructions
from invenio_db import db
from invenio_i18n import lazy_gettext as _

from .forms import EmailProfileForm, PreferencesForm, ProfileForm, VerificationForm
from .models import UserProfileProxy


def create_blueprint(app):
    """Create blueprint."""
    blueprint = Blueprint(
        "invenio_userprofiles",
        __name__,
        template_folder="templates",
        url_prefix="/account/settings",
    )

    @blueprint.app_template_filter()
    def userprofile(value):
        """Retrieve user profile for a given user id."""
        warn("userprofile template filter is deprecated.", DeprecationWarning)
        return UserProfileProxy.get_by_userid(int(value))

    blueprint.add_url_rule("/profile", view_func=profile, methods=["GET", "POST"])

    return blueprint


@login_required
def profile():
    """View for editing a profile."""
    # Create forms
    verification_form = VerificationForm(formdata=None, prefix="verification")
    profile_form = profile_form_factory()
    preferences_form = PreferencesForm(
        formdata=None, obj=current_user, prefix="preferences"
    )

    # Pick form
    is_read_only = current_app.config.get("USERPROFILES_READ_ONLY", False)
    form_name = request.form.get("submit", None)
    if form_name == "profile" and not is_read_only:
        handle_form = handle_profile_form
        form = profile_form
    elif form_name == "verification":
        handle_form = handle_verification_form
        form = verification_form
    elif form_name == "preferences":
        handle_form = handle_preferences_form
        form = preferences_form
    else:
        form = None

    # Process form
    if form:
        form.process(formdata=request.form)
        if form.validate_on_submit():
            handle_form(form)
            return redirect(url_for(".profile"), code=303)  # this endpoint

    return render_template(
        current_app.config["USERPROFILES_PROFILE_TEMPLATE"],
        verification_form=verification_form,
        profile_form=profile_form,
        preferences_form=preferences_form,
    )


def profile_form_factory():
    """Create a profile form."""
    if current_app.config["USERPROFILES_EMAIL_ENABLED"]:
        return EmailProfileForm(
            formdata=None,
            obj=current_user,
            prefix="profile",
        )
    else:
        return ProfileForm(
            formdata=None,
            obj=current_user,
            prefix="profile",
        )


def handle_verification_form(form):
    """Handle email sending verification form."""
    send_confirmation_instructions(current_user)
    # NOTE: Flash message.
    flash(_("Verification email sent."), category="success")


def handle_profile_form(form):
    """Handle profile update form."""
    email_changed = False
    datastore = current_app.extensions["security"].datastore
    with db.session.begin_nested():
        if (
            current_app.config["USERPROFILES_EMAIL_ENABLED"]
            and form.email.data.lower() != current_user.email.lower()
        ):
            email_changed = True
        form.populate_obj(current_user)
        db.session.add(current_user)
        datastore.mark_changed(id(db.session), uid=current_user.id)
    datastore.commit()

    if email_changed:
        send_confirmation_instructions(current_user)
        # NOTE: Flash message after successful update of profile.
        flash(
            _(
                "Profile was updated. We have sent a verification "
                "email to %(email)s. Please check it.",
                email=current_user.email.lower(),
            ),
            category="success",
        )
    else:
        # NOTE: Flash message after successful update of profile.
        flash(_("Profile was updated."), category="success")


def handle_preferences_form(form):
    """Handle preferences form."""
    form.populate_obj(current_user)
    db.session.add(current_user)
    current_app.extensions["security"].datastore.commit()
    # NOTE: Flash message after successful update of profile.
    flash(_("Preferences were updated."), category="success")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create userprofiles tables."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c25ef2c50ffa"
down_revision = "71634726ec7e"
branch_labels = ()
depends_on = "9848d0149abd"


def upgrade():
    """Upgrade database."""
    op.create_table(
        "userprofiles_userprofile",
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("username", sa.String(length=255), nullable=True),
        sa.Column("displayname", sa.String(length=255), nullable=True),
        sa.Column("full_name", sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["accounts_user.id"],
        ),
        sa.PrimaryKeyConstraint("user_id"),
        sa.UniqueConstraint("username"),
    )


def downgrade():
    """Downgrade database."""
    op.drop_table("userprofiles_userprofile")



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
# Copyright (C)      2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Remove table."""

import json

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "41157f1933d6"
down_revision = "c25ef2c50ffa"
branch_labels = ()
depends_on = "eb9743315a9d"  # invenio-accounts: add_userprofile


def upgrade():
    """Upgrade database."""
    # Migrate username and full_name to invenio_accounts table.
    connection = op.get_bind()
    results = connection.execute(
        text(
            "SELECT user_id, username, displayname, full_name "
            "FROM userprofiles_userprofile;"
        )
    ).all()

    for r in results:
        user_id, username, displayname, full_name = r
        profile_data = {
            "full_name": full_name,
        }
        preferences = {"visibility": "restricted", "email_visibility": "restricted"}

        query = (
            "UPDATE accounts_user SET "
            " username = '{username}', "
            " displayname = '{displayname}', "
            " profile = '{profile_data_string}', "
            " preferences = '{preferences}' "
            "WHERE accounts_user.id = {user_id};".format(
                username=username,
                displayname=displayname,
                profile_data_string=json.dumps(profile_data),
                preferences=json.dumps(preferences),
                user_id=user_id,
            )
        )
        op.execute(query)

    op.drop_table("userprofiles_userprofile")


def downgrade():
    """Downgrade database."""
    op.create_table(
        "userprofiles_userprofile",
        sa.Column("user_id", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column(
            "username", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "displayname", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "full_name", sa.VARCHAR(length=255), autoincrement=False, nullable=False
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["accounts_user.id"],
            name="fk_userprofiles_userprofile_user_id_accounts_user",
        ),
        sa.PrimaryKeyConstraint("user_id", name="pk_userprofiles_userprofile"),
        sa.UniqueConstraint("username", name="uq_userprofiles_userprofile_username"),
    )
